/*
 * rule.h -- rule for server and channels
 *
 * Copyright (c) 2013-2016 David Demelier <markand@malikania.fr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _IRCCD_RULE_H_
#define _IRCCD_RULE_H_

/**
 * @file Rule.h
 * @brief Rule description
 */

#include <sstream>
#include <string>
#include <unordered_set>
#include <utility>
#include <vector>

namespace irccd {

/**
 * List of criterias.
 */
using RuleSet = std::unordered_set<std::string>;

/**
 * @enum RuleAction
 * @brief Rule action
 */
enum class RuleAction {
	Accept,			//!< The event is accepted (default)
	Drop			//!< The event is dropped
};

/**
 * @class Rule
 * @brief Manage rule to activate or deactive events.
 */
class Rule final {
private:
	RuleSet m_servers;
	RuleSet m_channels;
	RuleSet m_origins;
	RuleSet m_plugins;
	RuleSet m_events;
	RuleAction m_action{RuleAction::Accept};

	/*
	 * Check if a map contains the value and return true if it is
	 * or return true if value is empty (which means applicable).
	 */
	bool matchMap(const RuleSet &map, const std::string &value) const noexcept;

public:
	/**
	 * Resolve the action to execute with the specified list of rules.
	 *
	 * @param rules the list of rules
	 * @param server the server name
	 * @param channel the channel name
	 * @param origin the origin
	 * @param plugin the plugin name
	 * @param event the event name (e.g onKick)
	 * @return true if the plugin must be called
	 */
	static bool solve(const std::vector<Rule> &rules,
			  const std::string &server,
			  const std::string &channel,
			  const std::string &origin,
			  const std::string &plugin,
			  const std::string &event) noexcept;

	/**
	 * Rule constructor.
	 *
	 * @param servers the server list
	 * @param channels the channels
	 * @param nicknames the nicknames
	 * @param plugins the plugins
	 * @param events the events
	 * @param action the rule action
	 * @throw std::invalid_argument if events are invalid
	 */
	Rule(RuleSet servers = RuleSet{},
	     RuleSet channels = RuleSet{},
	     RuleSet nicknames = RuleSet{},
	     RuleSet plugins = RuleSet{},
	     RuleSet events = RuleSet{},
	     RuleAction action = RuleAction::Accept);

	/**
	 * Check if that rule apply for the given criterias.
	 *
	 * @param server the server
	 * @param channel the channel
	 * @param nick the origin
	 * @param plugin the plugin
	 * @param event the event
	 * @return true if match
	 */
	bool match(const std::string &server,
		   const std::string &channel,
		   const std::string &nick,
		   const std::string &plugin,
		   const std::string &event) const noexcept;

	/**
	 * Get the action.
	 *
	 * @return the action
	 */
	RuleAction action() const noexcept;

	/**
	 * Get the servers.
	 *
	 * @return the servers
	 */
	const RuleSet &servers() const noexcept;

	/**
	 * Get the channels.
	 *
	 * @return the channels
	 */
	const RuleSet &channels() const noexcept;

	/**
	 * Get the origins.
	 *
	 * @return the origins
	 */
	const RuleSet &origins() const noexcept;

	/**
	 * Get the plugins.
	 *
	 * @return the plugins
	 */
	const RuleSet &plugins() const noexcept;

	/**
	 * Get the events.
	 *
	 * @return the events
	 */
	const RuleSet &events() const noexcept;
};

} // !irccd

#endif // !_IRCCD_RULE_H_
